/*
--
CubaseWindowBorderFIX v1.1
---

What is this application for?
--
This application fixes the following problem that occurs in Cubase 10 on Windows 10:
GUI shifted in some windows of Cubase 10 since the introduced window shadows in Windows 10
--

How does it achieve that?
--
Since the GUI in Cubase 10 is shown correctly after the window has been move, this application checks if a specified window was opened and moves it by 1 pixel to the right and back.
Only the windows that match the begining window title, specified in WindowTitleList.txt will be affected.
Depending on the language of your Cubase installation the WindowTitleList.txt needs to be edited to match the actual window titles.
This application was developped for Cubase 10, it probably works with v10.5 and also previous versions of Cubase.
--

Credits
---
v1.1 written by Frank Brempel in 2021
frank@blueingreen-recordings.de
https://www.blueingreen-recordings.de/CubaseWindowBorderFIX/

Icon made by Freepik from https://www.flaticon.com/

License
--
This software is open source and freeware.
*/

#include <Windows.h>
#include <windowsx.h>
#include <Knownfolders.h>
#include <Shlobj.h>
#include <stdio.h>
#include "resource.h"

#define VERSION L"v1.1"

#define WM_NOTIFYCALLBACK	WM_APP+1

#define ID_MENU_WT_LIST_EDIT	4001
#define ID_MENU_WT_LIST_RELOAD	4002
#define ID_MENU_CREDITS			4003
#define ID_MENU_CLOSE			4004
#define ID_MENU_SHOWHIDE		4005

#define CHECK_PARENT_WND_INTERVALL	10000 //ms
#define CHECK_CHILD_WND_INTERVALL	1000 //ms
#define MAX_WINDOW_TITLE_LIST	8

#define GUI_VPADDING		10
#define GUI_HPADDING		10
#define GUI_ENTRY_HEIGHT	18

class FIX_WINDOW
{
public:
	wchar_t title[64];
	bool is_fixed;
	HWND hwnd;

	FIX_WINDOW()
	{
		ZeroMemory(this, sizeof(FIX_WINDOW));
	}
};
FIX_WINDOW g_fix_window[MAX_WINDOW_TITLE_LIST];
int g_fix_window_length = 0;

HMENU g_hMenu = NULL;
HBRUSH g_brushDark;
HFONT g_fntNormal = NULL;

HWND g_hwndMain = NULL;
wchar_t g_window_text[256]=L"";

HWINEVENTHOOK g_hWinEventHook = NULL;

bool GetLocalAppDataPath(wchar_t *path)
{
	PWSTR appdata_path = NULL;
	if (SHGetKnownFolderPath(FOLDERID_LocalAppData, 0, NULL, &appdata_path) != S_OK)
		return false;

	swprintf_s(path, MAX_PATH, L"%s\\CubaseWindowBorderFIX", appdata_path);

	CoTaskMemFree(appdata_path);

	return true;
}

bool GetWindowTitleListPath(wchar_t *path)
{
	if (!GetLocalAppDataPath(path))
		return false;

	swprintf_s(path, MAX_PATH, L"%s\\WindowTitleList.txt", path);

	return true;
}

bool LoadWindowTitleList()
{
	FILE *fh;

	g_fix_window_length = 0;
	ZeroMemory(g_fix_window, sizeof(FIX_WINDOW)*MAX_WINDOW_TITLE_LIST);

	wchar_t file_path[MAX_PATH];
	if (!GetWindowTitleListPath(file_path))
		return false;

	errno_t result = _wfopen_s(&fh, file_path, L"r");
	if (result == ENOENT)
	{
		//datei nicht vorhanden -> erstellen
		wchar_t path[MAX_PATH];
		if (!GetLocalAppDataPath(path))
			return false;

		CreateDirectory(path,NULL);
		
		if (_wfopen_s(&fh, file_path, L"w") != 0)
			return false;

		fwprintf_s(fh, L"MixConsole\nPool\nVST-PlugIn-Manager\nAudio Connections");

		fclose(fh);

		if (_wfopen_s(&fh, file_path, L"r") != 0)
			return false;
	}
	else if(result != 0)
	{
		return false;
	}

	g_fix_window_length = 0;
	while (!feof(fh))
	{
		fgetws(g_fix_window[g_fix_window_length].title, 64, fh);
		for (int i = wcslen(g_fix_window[g_fix_window_length].title) -1 ; i > 0; i--)
		{
			if (g_fix_window[g_fix_window_length].title[i] == '\r'
				|| g_fix_window[g_fix_window_length].title[i] == '\n')
			{
				g_fix_window[g_fix_window_length].title[i] = '\0';
			}
			else
			{
				break;
			}
		}
		g_fix_window_length++;

		if (g_fix_window_length == MAX_WINDOW_TITLE_LIST)
			break;
	}

	fclose(fh);

	return true;
}

void SetCorrectWindowPosition(HWND hWnd, int x, int y, int width, int height)
{
	RECT rcClient;
	GetClientRect(hWnd, &rcClient);

	RECT rcWindow;
	GetWindowRect(hWnd, &rcWindow);

	int borders = ((rcWindow.right - rcWindow.left) - rcClient.right);
	int title_bar = ((rcWindow.bottom - rcWindow.top) - rcClient.bottom) - borders / 2;

	RECT rcWindowSize = { 0,0, width , height };
	AdjustWindowRect(&rcWindowSize, GetWindowLong(hWnd, GWL_STYLE), false);

	MoveWindow(hWnd, x + rcWindowSize.left, y + rcWindowSize.top,
		rcWindowSize.right - rcWindowSize.left, rcWindowSize.bottom - rcWindowSize.top, TRUE);
}

void CALLBACK WinEventProc(
	HWINEVENTHOOK hWinEventHook,
	DWORD event,
	HWND hwnd,
	LONG idObject,
	LONG idChild,
	DWORD idEventThread,
	DWORD dwmsEventTime
)
{
	if (hwnd &&
		idObject == OBJID_WINDOW &&
		idChild == CHILDID_SELF)
	{
		for (int i = 0; i < g_fix_window_length; i++)
		{
			if (event == EVENT_OBJECT_SHOW)
			{
				GetWindowText(hwnd, g_window_text, 256);

				if (_wcsnicmp(g_window_text, g_fix_window[i].title, wcslen(g_fix_window[i].title)) == 0)
				{
					if (!g_fix_window[i].is_fixed)
					{
						RECT rc;
						GetWindowRect(hwnd, &rc);
						MoveWindow(hwnd, rc.left + 1, rc.top + 1, rc.right - rc.left, rc.bottom - rc.top, FALSE);
						MoveWindow(hwnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

						g_fix_window[i].hwnd = hwnd;
						g_fix_window[i].is_fixed = true;

						RedrawWindow(g_hwndMain, NULL, NULL, RDW_INVALIDATE);
					}
					return;
				}
			}
			else if (event == EVENT_OBJECT_HIDE)
			{
				if (!g_fix_window[i].hwnd)
					continue;

				if (hwnd == g_fix_window[i].hwnd)
				{
					g_fix_window[i].is_fixed = false;
					g_fix_window[i].hwnd = NULL;

					RedrawWindow(g_hwndMain, NULL, NULL, RDW_INVALIDATE);
				}
			}
		}
	}
}

LRESULT WINAPI WindowProcMain(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:
	{
		g_brushDark = CreateSolidBrush(RGB(37, 40, 43));
		g_fntNormal = CreateFont(16, 0, 0, 0, FW_NORMAL, 0, 0, 0, DEFAULT_CHARSET,
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, L"Segoe UI");

		SendMessage(hWnd, WM_COMMAND, ID_MENU_WT_LIST_RELOAD, 0);

		return 0;
	}
	case WM_PAINT:
	{
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);

		SelectObject(hdc, g_fntNormal);
		SetBkMode(hdc, TRANSPARENT);
		SetTextColor(hdc, RGB(244,244,244));

		wchar_t text[1024] = L"";
		RECT rcClient;
		GetClientRect(hWnd, &rcClient);

		FillRect(hdc, &rcClient, (HBRUSH)g_brushDark);

		wchar_t state[8] = L"";

		for (int i = 0; i < g_fix_window_length; i++)
		{
			if (!wcslen(g_fix_window[i].title))
				continue;

			RECT rcLine = { GUI_HPADDING ,GUI_VPADDING + i * GUI_ENTRY_HEIGHT,
				rcClient.right - GUI_HPADDING, GUI_VPADDING + (i + 1) * GUI_ENTRY_HEIGHT};

			DrawText(hdc, g_fix_window[i].title, wcslen(g_fix_window[i].title), &rcLine, DT_LEFT | DT_VCENTER);

			wcscpy_s(state, 8, (g_fix_window[i].is_fixed) ? L"fixed" : L"closed");
			DrawText(hdc, state, wcslen(state), &rcLine, DT_RIGHT | DT_VCENTER);
		}

		EndPaint(hWnd, &ps);
		return 0;
	}
	case WM_RBUTTONUP:
	{
		int xPos = GET_X_LPARAM(lParam);
		int yPos = GET_Y_LPARAM(lParam);
		POINT pt = {xPos, yPos};
		ClientToScreen(hWnd, &pt);
		TrackPopupMenuEx(g_hMenu, 0, pt.x, pt.y, hWnd, NULL);
		return 0;
	}
	case WM_NOTIFYCALLBACK:
	{
		if (LOWORD(lParam) == NIN_SELECT)
		{
			ShowWindow(hWnd, SW_SHOW);
		}
		else if (LOWORD(lParam) == WM_CONTEXTMENU)
		{
			POINT pt = { LOWORD(wParam), HIWORD(wParam) };
			TrackPopupMenuEx(g_hMenu, TPM_CENTERALIGN | TPM_BOTTOMALIGN | TPM_VERPOSANIMATION, pt.x, pt.y, hWnd, NULL);
		}
		break;
	}
	case WM_COMMAND:
	{
		switch (wParam)
		{
		case ID_MENU_WT_LIST_EDIT:
		{
			wchar_t file_path[MAX_PATH];
			if (!GetWindowTitleListPath(file_path))
			{
				MessageBox(hWnd, L"Could not open 'WindowTitleList.txt'.", L"Error", MB_ICONERROR);
			}
			else
			{
				if ((int)ShellExecute(hWnd, L"edit", file_path, L"", L"", SW_SHOW) <= 32)
				{
					MessageBox(hWnd, L"Could not open 'WindowTitleList.txt'.", L"Error", MB_ICONERROR);
				}
			}
			break;
		}
		case ID_MENU_WT_LIST_RELOAD:
		{
			if (!LoadWindowTitleList())
			{
				MessageBox(hWnd, L"Could not open 'WindowTitleList.txt'.", L"Error", MB_ICONERROR);
			}

			RECT rcWorkArea;
			SystemParametersInfo(SPI_GETWORKAREA, 0, &rcWorkArea, 0);
			int wnd_width = 220 + 2 * GUI_HPADDING;
			int wnd_height = g_fix_window_length * GUI_ENTRY_HEIGHT + 2 * GUI_VPADDING;
			SetCorrectWindowPosition(hWnd, rcWorkArea.right - wnd_width, rcWorkArea.bottom - wnd_height, wnd_width, wnd_height);

			break;
		}
		case ID_MENU_CLOSE:
		{
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		}
		case ID_MENU_SHOWHIDE:
		{
			if (IsWindowVisible(hWnd))
				ShowWindow(hWnd, SW_HIDE);
			else
				ShowWindow(hWnd, SW_SHOW);
			break;
		}
		case ID_MENU_CREDITS:
		{
			if ((int)ShellExecute(hWnd, L"open", L"https://www.blueingreen-recordings.de/CubaseWindowBorderFIX/", L"", L"", SW_SHOW) <= 32)
			{
				MessageBox(hWnd, L"Could not open https://www.blueingreen-recordings.de/CubaseWindowBorderFIX/.", L"Error", MB_ICONERROR);
			}
			break;
		}
		}
		return 0;
	}
	case WM_SYSCOMMAND:
	{
		if (wParam == SC_CLOSE)
		{
			ShowWindow(hWnd, SW_HIDE);
			return 0;
		}
		break;
	}
	case WM_DESTROY:
	{
		DeleteObject(g_brushDark);
		g_brushDark = NULL;

		DeleteObject(g_fntNormal);
		g_fntNormal = NULL;

		PostQuitMessage(0);
		return 0;
	}
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


HWND CreateMainWindow(HINSTANCE hInstance, const wchar_t *name)
{
	WNDCLASS wc;
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wc.lpfnWndProc = WindowProcMain;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName = name;
	wc.lpszClassName = L"CubaseWindowBorderFIX";
	RegisterClass(&wc);

	HWND hwndMain = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_TOPMOST, L"CubaseWindowBorderFIX", name, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU, 0, 0, 180, 100,
		NULL, NULL, hInstance, NULL);
	if (!hwndMain)
		return NULL;

	return hwndMain;
}

void ReleaseMainWindow(HWND hwnd, HINSTANCE hInstance)
{
	DestroyWindow(hwnd);
	UnregisterClass(L"CubaseWindowBorderFIX", hInstance);
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, char *lpCmdLine, int nCmdShow)
{
	MSG msg;
	HANDLE hMutex = CreateMutex(NULL, FALSE, L"CubaseWindowBorderFIX");
	if (hMutex && GetLastError() == ERROR_ALREADY_EXISTS)
	{
		//app already running
		HWND hwndRunning = FindWindow(L"CubaseWindowBorderFIX", NULL);
		ShowWindow(hwndRunning, SW_SHOW);
		return 0;
	}

	g_hWinEventHook = SetWinEventHook(
		EVENT_OBJECT_SHOW, EVENT_OBJECT_HIDE,
		NULL, (WINEVENTPROC)WinEventProc, 0, 0,
		WINEVENT_OUTOFCONTEXT | WINEVENT_SKIPOWNPROCESS);

	if (!g_hWinEventHook)
	{
		MessageBox(NULL, L"SetWinEventHook failed.", L"Error", MB_ICONERROR);
		return 0;
	}

	wchar_t wnd_title[64];
	swprintf_s(wnd_title, 64, L"CubaseWindowBorderFIX %s", VERSION);
	g_hwndMain = CreateMainWindow(hInstance, wnd_title);
	if (g_hwndMain)
	{
		NOTIFYICONDATA nid;
		nid.cbSize = sizeof(NOTIFYICONDATA);
		nid.hWnd = g_hwndMain;
		nid.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
		nid.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
		nid.uVersion = NOTIFYICON_VERSION_4;
		nid.uCallbackMessage = WM_NOTIFYCALLBACK;
		wcscpy_s(nid.szTip, 128 , L"CubaseWindowBorderFIX");
		Shell_NotifyIcon(NIM_ADD, &nid);
		Shell_NotifyIcon(NIM_SETVERSION, &nid);

		g_hMenu = CreatePopupMenu();
		MENUITEMINFO mii = { 0 };
		mii.cbSize = sizeof(MENUITEMINFO);

		wchar_t menu_txt[64];
		swprintf_s(menu_txt, 64, L"About %s by Frank Brempel", VERSION);
		mii.fMask = MIIM_STRING | MIIM_ID;
		mii.dwTypeData = (LPWSTR)menu_txt;
		mii.cch = wcslen(mii.dwTypeData);
		mii.wID = ID_MENU_CREDITS;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_TYPE;
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_STRING | MIIM_ID;
		mii.dwTypeData = (LPWSTR)L"Close";
		mii.cch = wcslen(mii.dwTypeData);
		mii.wID = ID_MENU_CLOSE;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_STRING | MIIM_ID;
		mii.dwTypeData = (LPWSTR)L"Show/hide window";
		mii.cch = wcslen(mii.dwTypeData);
		mii.wID = ID_MENU_SHOWHIDE;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_TYPE;
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_STRING | MIIM_ID;
		mii.dwTypeData = (LPWSTR)L"Edit WindowTitleList.txt";
		mii.cch = wcslen(mii.dwTypeData);
		mii.wID = ID_MENU_WT_LIST_EDIT;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		mii.fMask = MIIM_STRING | MIIM_ID;
		mii.dwTypeData = (LPWSTR)L"Reload WindowTitleList.txt";
		mii.cch = wcslen(mii.dwTypeData);
		mii.wID = ID_MENU_WT_LIST_RELOAD;
		InsertMenuItem(g_hMenu, 0, TRUE, &mii);

		if (strcmp(lpCmdLine, "hidden") == 0)
			ShowWindow(g_hwndMain, SW_HIDE);
		else
			ShowWindow(g_hwndMain, SW_SHOW);

		while (GetMessage(&msg, NULL, 0, 0))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		DestroyMenu(g_hMenu);

		Shell_NotifyIcon(NIM_DELETE, &nid);

		ReleaseMainWindow(g_hwndMain, hInstance);
	}

	UnhookWinEvent(g_hWinEventHook);
	g_hWinEventHook = NULL;

	CloseHandle(hMutex);

	return 0;
}